angular.module('beamng.apps')
  .directive('dashcamui', [function () {
    return {
      templateUrl: '/ui/modules/apps/dashcamui/app.html',
      replace: true,
      restrict: 'EA',
      link: function ($scope, element, attrs) {
        const appName = "Dashcam Overlay UI"
        const appVer = "v0.2"
        
        //streams
        var streamsList = ['electrics', 'sensors'];
        StreamsManager.add(streamsList);

        //used for getting or setting values of elements in options window
        const elements = {
          set fontSize(value) { document.getElementById('fontSize').value = value },
          get fontSize() { return document.getElementById('fontSize').value },
          set date(value) { document.getElementById('enterDate').value = value },
          get date() { return document.getElementById('enterDate').value },
          set gpsDelay(value) { document.getElementById('GPSdelay').value = value },
          get gpsDelay() { return document.getElementById('GPSdelay').value },
          set showampm(value) { document.getElementById('ampm').checked = value },
          get showampm() { return document.getElementById('ampm').checked },
          set showSpeed(value) { document.getElementById('speed').checked = value },
          get showSpeed() { return document.getElementById('speed').checked },
          set showCoords(value) { document.getElementById("coords").checked = value },
          get showCoords() { return document.getElementById('coords').checked },
          set showGForce(value) { document.getElementById('gforce').checked = value },
          get showGForce() { return document.getElementById('gforce').checked },
          set orangeColor(value) { document.getElementById('orange').checked = value },
          get orangeColor() { return document.getElementById('orange').checked },
          set autoRecordMinG(value) { document.getElementById('minG').value = value },
          get autoRecordMinG() { return document.getElementById('minG').value },
          set audioNum(value) { document.getElementById('audio').value = value },
          get audioNum() { return document.getElementById('audio').value },
          set speechNum(value) { document.getElementById('speech').value = value },
          get speechNum() { return document.getElementById('speech').value },
          set dashcamModel(value) {
            switch (value) {
              case 'Hide':
                document.getElementById('model').value = '0'
                break;
              case 'Flyer':
                document.getElementById('model').value = '2'
                break;
              default:
                document.getElementById('model').value = '1'
                break;
            }
          },
          get dashcamModel() {
            switch (document.getElementById('model').value) {
              case '0':
                return 'Hide';
              case '2':
                return 'Flyer';
              default:
                return 'BeamCam';
            }
          }
        }
        //config values
        let fontSize = 25
        let gpsDelay = 2
        let showampm = 0
        let showSpeed = 1
        let showCoords = 1
        let showGForce = 0
        let orangeColor = 0
        let autoRecordMinG = 3
        //$scope.options = {
        //  dashcamModel: ['Hide', 'BeamCam', 'Flyer'],
        //  audioNum: ['No beep', '1', '2', '3', '4'],
        //  speechNum: ['No speech', '1', '2']
        //};

        //let audioNum = $scope.getIndex($scope.audioNum, $scope.options.audioNum);
        //let speechNum = $scope.getIndex($scope.speechNum, $scope.options.speechNum);
        //let dashcamModel = $scope.dashcamModel;
        let audioNum = 0
        let speechNum = 0
        let dashcamModel = 'BeamCam'

        //files
        const filesFolder = '/ui/modules/apps/dashcamui/files/'
        var audio
        var speech
        var testing = new Audio(filesFolder + 'testing.mp3')

        //values
        let audioPlaying = false
        let audioEnabled = false

        let date
        let hours = 0
        let minutes = -1
        let seconds = -1

        let gpsSpeed = 0
        let count = 0
        let sum = 0

        let longitude
        let latitude

        let maxGForce = 0 //max GForce recorded in beetween overlay updates
        let gForce = 0 //current value displayed

        element.ready(function () { init() });

        function init() {
          document.getElementById('title').textContent = `${appName} ${appVer} // Options`
          //get local date
          date = elements.date = getLocalDate()

          //get in-game time for clock
          getGameTime()

          //random coordinates to start with
          latitude = (Math.random() * 180) - 90;
          longitude = (Math.random() * 360) - 180;

          //reading and applying config
          readUserPref()
          applyConfig()

          //animation on startup
          document.getElementById("options").classList.add('welcome')

          //click event for showing options, or enabling audio
          document.getElementById("logo").addEventListener("click", function () {
            if ((audioNum > 0 || speechNum > 0) && !audioEnabled) { applyConfig() }
            else { document.getElementById("options").classList.remove('welcome'); document.getElementById("options").classList.add('options-open') }
          })

          //click events for various elements in options window
          document.getElementById("btnApply").addEventListener("click", function () {
            applyOptions(); saveUserPref();
          })
          document.getElementById("resetDate").addEventListener("click", function () {
            date = elements.date = getLocalDate()
          })
          document.getElementById("model").oninput = document.getElementById("audio").oninput = document.getElementById("speech").oninput = sliderUpdate;

          //update slider appearance after setting up config
          sliderUpdate(document.getElementById('audio'), false)
          sliderUpdate(document.getElementById('speech'), false)
          sliderUpdate(document.getElementById('model'), false)
        }

        function setupAudioandDisplay() {
          resetAudio()
          if (audioNum > 0 || speechNum > 0) {
            if (audioNum > 0 && (!audio || audio.src != filesFolder + audioNum + '.mp3')) { audio = new Audio(filesFolder + audioNum + '.mp3') }
            if (speechNum > 0 && (!speech || speech.src != filesFolder + "speech" + speechNum + '.mp3')) { speech = new Audio(filesFolder + "speech" + speechNum + '.mp3') }
            if (audioEnabled) { applyDashcamModel(); return }
            testing.play().then(function () { //solution to play() failed because the user didn't interact with the document first
              audioEnabled = true
              //log("Audio Enabled", 'info');
              applyDashcamModel() //set logo and most of config options after the message
            }).catch(function (error) {
              log("Audio Disabled; " + error.message);
              audioEnabled = false;
              document.getElementById("logo").style.height = '0.7em'
              document.getElementById("logo").src = filesFolder + 'off.png';
              document.getElementById("logo-text").textContent = "Click the button to enable audio; " + error.message;
            });
          } else { applyDashcamModel() } //no message needed, proceed to applying config
        }
        function resetAudio() {
          if (audio) { audio.pause(); audio.currentTime = 0; }
          if (speech) { speech.pause(); speech.currentTime = 0; }
          audioPlaying = false
        }
        function applyDashcamModel() {
          document.getElementById("logo").style.opacity = dashcamModel === 'Hide' ? 0 : 1
          document.getElementById("overlay").style.color = orangeColor ? "#ff6600" : "#ffffff"
          switch (dashcamModel) {
            case 'BeamCam':
              document.getElementById("overlay").style.fontFamily = 'Roboto'
              document.getElementById("logo").style.height = '0.5em'
              document.getElementById("logo").src = orangeColor ? filesFolder + "BeamNG-logo-icon-2016.svg" : filesFolder + "BeamNG-logo-icon-2016-white.svg"
              document.getElementById("logo-text").textContent = 'BeamCam'
              break;
            case 'Flyer':
              document.getElementById("overlay").style.fontFamily = 'hemi'
              document.getElementById("logo").style.height = '1em'
              document.getElementById("logo").src = orangeColor ? filesFolder + "flyer-orange.png" : filesFolder + "flyer.png"
              //document.getElementById("logo").style.filter = orangeColor ? 'brightness(50%) sepia(100) saturate(100) hue-rotate()' : ""
              //document.getElementById("logo").style.boxShadow = '0 4px 8px 0 rgba(0, 0, 0, 0.2), 0 6px 20px 0 rgba(0, 0, 0, 0.19)'
              document.getElementById("logo-text").textContent = '\xa0\xa0Flyer' //using non-braking space due to logo being more 'filled'
              break;
            default:
              document.getElementById("overlay").style.fontFamily = 'vcr'
              document.getElementById("logo").style.height = '0.5em'
              document.getElementById("logo").src = orangeColor ? filesFolder + "BeamNG-logo-icon-2016.svg" : filesFolder + "BeamNG-logo-icon-2016-white.svg"
              document.getElementById("logo-text").textContent = ''
              break;
          }
        }
        //getting values from options window, then calling applyConfig()
        function applyOptions() {
          date = elements.date;
          gpsDelay = elements.gpsDelay;
          showampm = elements.showampm;
          showSpeed = elements.showSpeed;
          showCoords = elements.showCoords;
          showGForce = elements.showGForce;
          orangeColor = elements.orangeColor;
          dashcamModel = elements.dashcamModel;
          autoRecordMinG = elements.autoRecordMinG;
          audioNum = elements.audioNum;
          speechNum = elements.speechNum;
          fontSize = elements.fontSize
          applyConfig()
          document.getElementById("options").classList.remove('options-open') //closing options window
        }
        function sliderUpdate(element, testPlay = true) {
          if (typeof element === 'object' && element instanceof Event) {
            element = element.target; //if the function is called from GlobalEventHandlers.oninput, convert Event to element
          }
          if (element.id === 'model') {
            switch (element.value) {
              case '0' || 0:
                document.getElementById("output").innerHTML = 'Professional';
                break;
              case '2' || 2:
                document.getElementById("output").innerHTML = "Flyer";
                break;
              default:
                document.getElementById("output").innerHTML = "BeamCam";
                break;
            }
          }
          else if (element.id === 'audio' || element.id === 'speech') {
            if (element.value === '0') {
              element.classList.add('sliderOFF'); //making handle darker if turned OFF
            } else {
              element.classList.remove('sliderOFF');
              if (!testPlay) { return } //dont play audio when loading config
              const isAudio = element.id === 'audio';
              const fileName = isAudio ? element.value : 'speech' + element.value;
              const testAudio = new Audio(filesFolder + fileName + '.mp3');
              testAudio.currentTime = isAudio ? 0.5 : 1.5;
              testAudio.play(); //play audio so that user can hear what it sounds like
            }
          }
        }
        //updating few options
        function applyConfig() {
          //document.getElementById("overlay").style.fontFamily = fontName;
          if (fontSize < 10) { fontSize = 10 }
          document.getElementById("overlay").style.fontSize = fontSize / 10 + "em";
          let shadow = fontSize / 10 - 1
          document.getElementById("overlay").style.textShadow = `${-shadow}px ${-shadow}px 0 #000, ${shadow}px ${-shadow}px 0 #000, ${-shadow}px ${shadow}px 0 #000, ${shadow}px ${shadow}px 0 #000`;
          setupAudioandDisplay() //updating the rest
        }
        function readUserPref() {
          let value = localStorage.getItem('dashcamUI-options');
          //log(value, 'info')
          let settings
          if (value) { settings = JSON.parse(value); }
          else { settings = {} } //no config detected, assigning default values
          elements.date = settings.hasOwnProperty('date') ? settings.date : date;
          elements.gpsDelay = !isNaN(settings.gpsDelay) ? settings.gpsDelay : gpsDelay;
          elements.showampm = settings.hasOwnProperty('showampm') ? settings.showampm : showampm;
          elements.showSpeed = settings.hasOwnProperty('showSpeed') ? settings.showSpeed : showSpeed;
          elements.showCoords = settings.hasOwnProperty('showCoords') ? settings.showCoords : showCoords;
          elements.showGForce = settings.hasOwnProperty('showGForce') ? settings.showGForce : showGForce;
          elements.orangeColor = settings.hasOwnProperty('orangeColor') ? settings.orangeColor : orangeColor;
          elements.dashcamModel = settings.hasOwnProperty('dashcamModel') ? settings.dashcamModel : dashcamModel;
          elements.autoRecordMinG = !isNaN(settings.autoRecordMinG) ? settings.autoRecordMinG : autoRecordMinG;
          elements.audioNum = !isNaN(settings.audioNum) ? settings.audioNum : audioNum;
          elements.speechNum = !isNaN(settings.speechNum) ? settings.speechNum : speechNum;
          elements.fontSize = !isNaN(settings.fontSize) ? settings.fontSize : fontSize;
          applyOptions() //update elements in options window
        }
        function saveUserPref() {
          var options = {
            date: getLocalDate() === date ? undefined : date, //if date is not changed (default), make sure not to save it so that the next day a correct one is displayed
            gpsDelay: gpsDelay,
            fontSize: fontSize,
            showSpeed: showSpeed,
            showCoords: showCoords,
            showGForce: showGForce,
            showampm: showampm,
            orangeColor: orangeColor,
            autoRecordMinG: autoRecordMinG,
            audioNum: audioNum, //0 - disabled, 1-4 - enabled
            speechNum: speechNum, //0 - disabled, 1-2 - enabled
            dashcamModel: dashcamModel
            //fontName: fontName
          };
          localStorage.setItem('dashcamUI-options', JSON.stringify(options));
          //log(localStorage.getItem('dashcamUI-options'), 'info')
        }

        function calculateGForce(streams) {
          var gravity = streams.sensors.gravity
          gravity = gravity >= 0 ? Math.max(0.1, gravity) : Math.min(-0.1, gravity)
          let xg = Math.abs(streams.sensors.gx2) / -gravity;
          let yg = Math.abs(streams.sensors.gy2) / -gravity;
          let result = (xg + yg);
          //logic for getting the highest reading beetween overlay updates
          if (result > maxGForce) { maxGForce = result }
          if (result >= autoRecordMinG) { autoRecord() }
        }
        function calculateNewCoordinates(startLat, startLon, speed, directionDegrees, intervalMillis) {
          // Radius of the Earth in meters
          const R = 6378137;

          // Convert direction from degrees to radians
          const directionRadians = directionDegrees * Math.PI / 180;

          // Calculate the distance traveled in the given time interval
          const distance = speed / 10 * intervalMillis;

          // Calculate the new latitude using the Haversine formula
          const newLat = Math.asin(
            Math.sin(startLat) * Math.cos(distance / R) +
            Math.cos(startLat) * Math.sin(distance / R) * Math.cos(directionRadians)
          );

          // Calculate the new longitude using the Haversine formula
          const newLon = startLon + Math.atan2(
            Math.sin(directionRadians) * Math.sin(distance / R) * Math.cos(startLat),
            Math.cos(distance / R) - Math.sin(startLat) * Math.sin(newLat)
          );

          return [newLat, newLon];
        }
        function autoRecord() {
          if (audioNum > 0 || speechNum > 0) {
            if (audioEnabled) {
              if (!audioPlaying) {
                if (audioNum > 0) {
                  audioPlaying = true
                  audio.play()
                  audio.addEventListener("ended", function () { audioPlaying = false; });
                  if (speechNum > 0) { speech.play() }
                }
                else if (speechNum > 0) {
                  audioPlaying = true
                  speech.currentTime = 1
                  speech.play()
                  speech.addEventListener("ended", function () { audioPlaying = false; });
                }
              }
            } else { log("Audio failed because the user didn't interact with the document first") }
          }
        }
        function checkForUpdate() {
          let currentDate = new Date();
          if (currentDate.getSeconds() != seconds) {
            seconds = currentDate.getSeconds();
            if (seconds === 0) {
              minutes++
              if (minutes === 60) {
                minutes = 0;
                hours++;
                if (hours >= 24) {
                  hours = 0;
                }
              }
            }
            if (gpsDelay <= 1) { return true }
            return seconds % gpsDelay === 0
          }
          return gpsDelay === "0"
        }

        function getData(streams) {
          if (streams === undefined) { return null } //showing logo only

          sum += streams.electrics.airspeed
          count++

          if (checkForUpdate()) { //simulating gps refresh delay
            gpsSpeed = sum / count; //calculating avg speed
            [latitude, longitude] = calculateNewCoordinates(latitude, longitude, gpsSpeed, streams.sensors.yaw * 180 / Math.PI + 180, Math.max(1000 / 30, gpsDelay * 1000))

            gForce = maxGForce; //showing max recorded G
            sum = count = maxGForce = 0; //resetting
          }

          let res = UiUnits.speed(gpsSpeed) //speed

          res.val = showSpeed ? String(res.val.toFixed()).padStart(3, '0') + res.unit : "" //formatted speed + unit

          res.coords = showCoords ? formatCoordinates(latitude, longitude) : ""

          res.date = date
          res.time = showampm ? res.time = convertTo12H(hours, minutes) : String(hours).padStart(2, '0') + ":" + String(minutes).padStart(2, '0') + ":" + String(seconds).padStart(2, '0')

          if (showGForce || audioNum > 0 || speechNum > 0) { calculateGForce(streams) } //calculate only if necessary
          res.gforce = showGForce ? gForce.toFixed(1) + "G" : ""

          return res
        }

        //reset values when needed
        $scope.$on('VehicleFocusChanged', function () {
          sum = count = gpsSpeed = gForce = maxGForce = 0;
          resetAudio()
        })
        $scope.$on('VehicleReset', function () {
          sum = count = gpsSpeed = gForce = maxGForce = 0;
          resetAudio()
        })
        $scope.$on('VehicleChange', function () {
          sum = count = gpsSpeed = gForce = maxGForce = 0;
          resetAudio()
        })

        //clean up after closing the app
        $scope.$on('$destroy', function () {
          StreamsManager.remove(streamsList);
          $scope.stop;
        });
        //updating
        $scope.$on('streamsUpdate', function (event, data) {
          $scope.set = getData(data)
          $scope.$apply()
        })

        //utilities
        function log(msg, severity = 'warn') {
          switch (severity) {
            case 'info':
              console.log(appName + ": " + msg); break;
            case 'warn':
              console.warn(appName + ": " + msg); break;
            case 'error':
              console.error(appName + ": " + msg); break;
          }
        }
        function getLocalDate() {
          let currentDate = new Date();
          let year = currentDate.getFullYear();
          let month = String(currentDate.getMonth() + 1).padStart(2, '0');
          let day = String(currentDate.getDate()).padStart(2, '0');
          return `${year}-${month}-${day}`
        }
        function getGameTime() {
          bngApi.engineLua('core_environment.getTimeOfDay().time', function (data) {
            //data - actual time between ''0'' and ''1'' where ''0'' and ''1'' are noon (12pm) and 0.5 is midnight (12am)
            data = (12 + data * 24) % 24
            hours = Math.floor(data)
            minutes = Math.floor((data - hours) * 60)
          });
        }
        function convertTo12H(hour, minute) {
          let period = hour >= 12 ? ' PM' : ' AM'
          hour = hour % 12;
          hour = hour ? hour : 12;

          hour = String(hour).padStart(2, '0')
          minute = String(minute).padStart(2, '0')
          seconds = String(seconds).padStart(2, '0')

          return hour + ':' + minute + ':' + seconds + ' ' + period;
        }
        function formatCoordinates(latitude, longitude) {
          // Convert latitude to degrees, minutes, seconds, and direction
          var latDegrees = Math.floor(Math.abs(latitude));
          var latMinutes = Math.floor((Math.abs(latitude) - latDegrees) * 60);
          var latSeconds = Math.round(((Math.abs(latitude) - latDegrees) * 3600) % 60);
          var latDirection = latitude >= 0 ? "N" : "S";

          // Convert longitude
          var lonDegrees = Math.floor(Math.abs(longitude));
          var lonMinutes = Math.floor((Math.abs(longitude) - lonDegrees) * 60);
          var lonSeconds = Math.round(((Math.abs(longitude) - lonDegrees) * 3600) % 60);
          var lonDirection = longitude >= 0 ? "E" : "W";

          // Format the coordinates string and return ready-to-print text
          return (String(latDegrees).padStart(3, '0') + "°" +
            String(latMinutes).padStart(3, '0') + "'" +
            String(latSeconds).padStart(3, '0') + "\"" +
            latDirection + " " +
            String(lonDegrees).padStart(3, '0') + "°" +
            String(lonMinutes).padStart(3, '0') + "'" +
            String(lonSeconds).padStart(3, '0') + "\"" +
            lonDirection);
        }
      }
    }
  }])